import setuptools

setuptools.setup(
    name='mrsketch',
    version='0.0.2',
    description='Simple library for building multi-dimensional multi-resolution sketches',
    url='https://github.com/Simonl07/mrsketch',
    author='Simon Lu',
    author_email='mlu18@dons.usfca.edu',
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3',
    ]
)
